#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#ifdef JFB_ENABLE_SCROLLBACK
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include "fbcommon.h"
#include "font.h"
#include "vterm.h"
/*
 * retrieve histlen size
 */
#define HISTLEN_DEFAULT 256
static int histlen = 0;
void scrollback_set_histlen(int len)
{
	if(0 < len && len < 10000)
		histlen = len;
	else
		histlen = HISTLEN_DEFAULT;
}

int scrollback_get_histlen(void)
{
	return histlen;
}

/*
 * keybord manipulation
 */
#include <linux/kd.h>
#include <linux/keyboard.h>

#define SHIFT_TABLE 1
#define PAGE_UP_KEYCODE 104
#define PAGE_DOWN_KEYCODE 109
#define PAGE_UP_ACTION 0x0118
#define PAGE_DOWN_ACTION 0x0119

u_short scroll_forward = 0;
u_short scroll_backward = 0;

int scrollback_set_kb(u_short table, u_short keycode,
		      u_short value, u_short *oldvalue, u_short readonly)
{
	struct kbentry ke;
	ke.kb_table = table;
	ke.kb_index = keycode;

	if(ioctl(fileno(stdin), KDGKBENT, &ke))
		return -1;

	if(readonly == 0 && oldvalue)
		*oldvalue = ke.kb_value;

	ke.kb_value = value;

	if(ioctl(fileno(stdin), KDSKBENT, &ke))
		return -1;

	return 0;
}

void scrollback_restore_key_atexit(void)
{
	if(scroll_backward)
		scrollback_set_kb(SHIFT_TABLE, PAGE_UP_KEYCODE,
				  scroll_backward, 0, 0);
	if(scroll_forward)
		scrollback_set_kb(SHIFT_TABLE, PAGE_DOWN_KEYCODE,
				  scroll_forward, 0, 0);
}

void scrollback_restore_key(void)
{
	if(scroll_backward)
		scrollback_set_kb(SHIFT_TABLE, PAGE_UP_KEYCODE,
				  scroll_backward, 0, 1);
	if(scroll_forward)
		scrollback_set_kb(SHIFT_TABLE, PAGE_DOWN_KEYCODE,
				  scroll_forward, 0, 1);
}

int scrollback_disable_key(u_short readonly)
{
	if(scrollback_set_kb(SHIFT_TABLE, PAGE_UP_KEYCODE,
			     PAGE_UP_ACTION, &scroll_backward, readonly))
		return -1;
	if(scrollback_set_kb(SHIFT_TABLE, PAGE_DOWN_KEYCODE,
			     PAGE_DOWN_ACTION, &scroll_forward, readonly))
		return -1;
	atexit(scrollback_restore_key_atexit);
	return 0;
}

/*
 * scrollback memory management
 */
static inline u_int scrollback_coord_to_index(TVterm *p, u_int x, u_int y)
{
	return (p->textHead + x + y * p->xcap4) % (p->xcap4 * p->histsize);
}

void scrollback_move_to_hist(TVterm *p, int dst, int src, int n)
{
	memmove(p->htext+dst, p->text+src, n*sizeof(u_int));
	memmove(p->hattr+dst, p->attr+src, n);
	memmove(p->hflag+dst, p->flag+src, n);
}

void scrollback_move_from_hist(TVterm *p, int dst, int src, int n)
{
	memmove(p->text+dst, p->htext+src, n*sizeof(u_int));
	memmove(p->attr+dst, p->hattr+src, n);
	memmove(p->flag+dst, p->hflag+src, n);
}

/*
 * crippled circular queue
 */
static int queue_front = -1;
static int queue_rear = 0;

int queue_get_size(TVterm *p)
{
	if(queue_front == -1)
		return 0;
	else if(queue_front == queue_rear)
		return p->histsize;
	return queue_rear - queue_front;
}

int queue_line_rel_to_abs(TVterm *p, int rel)
{
	u_int queue_size = queue_get_size(p);
	if(0 <= rel && rel < queue_size)
		return (queue_rear - (rel + 1) + p->histsize) % p->histsize;
	return -1;
}

void queue_insert_line(TVterm *p, int src)
{
	int dst, queue_size;
	dst = scrollback_coord_to_index(p, 0, queue_rear);
	scrollback_move_to_hist(p, dst, src, p->xcap4);

	queue_size = queue_get_size(p);
	if(queue_size == 0 || queue_size == p->histsize)
		queue_front = ++queue_front % p->histsize;
	queue_rear = ++queue_rear % p->histsize;
}

/*
 * scrollback refresh routine
 */
void scrollback_refresh(TVterm *p, int page)
{
	int page_front, page_rear, queue_size;
	int i, iy, x, y;

	u_int w, gw, chlw, lang;
	u_char bc, fc, fg, ch, ch2;
	TFont *pf;
	const u_char *glyph;

	tvterm_hide_cursor(p);
	page_front = (p->ycap / 2) * page;
	queue_size = queue_get_size(p);
	if(page_front > queue_size)
		page_front = queue_size;
	page_rear = page_front - p->ycap;
	if(page_rear == 0) page_rear = 0;

	for(iy = page_front - 1; iy >= page_rear; iy--) {
		y = queue_line_rel_to_abs(p, iy);
		for(x = 0; x < p->xcap; x++) {
			i = scrollback_coord_to_index(p, x, y);
			p->hflag[i] = p->hflag[i] & ~CLEAN_S;
		}
	}

	for(iy = page_front - 1; iy >= page_rear; iy--) {
		y = queue_line_rel_to_abs(p, iy);
		for(x = 0; x < p->xcap; x++) {
			i = scrollback_coord_to_index(p, x, y);
			w = 1;
			fg = p->hflag[i];
			if(fg & CLEAN_S)
				continue;
			fc = p->hattr[i] & 0xf;
			bc = p->hattr[i] >> 4;
			chlw = p->htext[i];
			ch2 = (chlw >> 8) & 0xff;
			ch = chlw & 0xff;
			lang = (chlw >> 24) & 0xff;
			p->hflag[i] |= CLEAN_S;
			if(fg & CODEIS_1) {
				pf = &(gFont[lang]);
				i++;
				p->hflag[i] |= CLEAN_S;
				w = 2;
			}
			else
				pf = &(gFont[lang]);
			glyph = pf->conv(pf, chlw, &gw);
			gFramebuffer.cap.fill(&gFramebuffer,
				gFontsWidth * x, gFontsHeight * (page_front - 1 - iy),
				gFontsWidth * w, gFontsHeight, bc);
			if(chlw == 0)
				continue;
			gFramebuffer.cap.overlay(&gFramebuffer,
				gFontsWidth * x, gFontsHeight * (page_front - 1 - iy),
				glyph, gw, pf->height, pf->bytew, fc);
		}
	}

	for(y = 0; y < (p->ycap - page_front); y++) {
		for(x = 0; x < p->xcap; x++) {
			i = scrollback_coord_to_index(p, x, y);
			p->flag[i] = p->flag[i] & ~CLEAN_S;
		}
	}

	for(y = 0; y < (p->ycap - page_front); y++) {
		for(x = 0; x < p->xcap; x++) {
			i = scrollback_coord_to_index(p, x, y);
			w = 1;
			fg = p->flag[i];
			if(fg & CLEAN_S)
				continue;
			fc = p->attr[i] & 0xf;
			bc = p->attr[i] >> 4;
			chlw = p->text[i];
			ch2 = (chlw >> 8) & 0xff;
			ch = chlw & 0xff;
			lang = (chlw >> 24) & 0xff;
			p->flag[i] |= CLEAN_S;
			if(fg & CODEIS_1) {
				pf = &(gFont[lang]);
				i++;
				p->flag[i] |= CLEAN_S;
				w = 2;
			}
			else
				pf = &(gFont[lang]);
			glyph = pf->conv(pf, chlw, &gw);
			gFramebuffer.cap.fill(&gFramebuffer,
				gFontsWidth * x, gFontsHeight * (y + page_front),
				gFontsWidth * w, gFontsHeight, bc);
			if(chlw == 0)
				continue;
			gFramebuffer.cap.overlay(&gFramebuffer,
				gFontsWidth * x, gFontsHeight * (y + page_front),
				glyph, gw, pf->height, pf->bytew, fc);
		}
	}

	for(y = 0; y < p->ycap; y++) {
		for(x = 0; x < p->xcap; x++) {
			i = scrollback_coord_to_index(p, x, y);
			p->flag[i] = p->flag[i] & ~CLEAN_S;
		}
	}
}

/*
 * scrollback actions
 */
static int pageno = 0;
void scrollback_init_pageno(void)
{
	pageno = 0;
}

void scrollback_scroll_up(TVterm *p)
{
	int max_scroll_pages;
	max_scroll_pages = queue_get_size(p) / (p->ycap / 2);
	if(max_scroll_pages >= pageno)
		scrollback_refresh(p, ++pageno);
}

void scrollback_scroll_down(TVterm *p)
{
	if(pageno <= 0)
		tvterm_refresh(p);
	else
		scrollback_refresh(p, --pageno);
}

int scrollback_is_on(void)
{
	if(pageno)
		return 1;
	return 0;
}
#endif
